package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class constructs a DFA from an NFA
 */
class DFA implements Graph, NFA
{
    /** The NFA to construct from */
    NFA nfa;

    /** NFA start states */
    Collection roots;

    /** Map from sets of NFA states to DFA states */
    Map dfaStates;

    /** DFA states */
    List nodes;

    /** DFA transitions */
    Map edges;

    /** Map from NFA start state to DFA start state. */
    DFAState[] startStates;

    /** Map from NFA accept state to DFA accept state. */
    BitSet[] acceptStates;

    /**
     * Build a DFA for the NFA.
     * To avoid rebuilding the DFA, the NFA and DFA can have multiple start
     * states.
     */
    DFA(NFA nfa) {
        this(nfa, Collections.singleton(nfa.start()));
    }

    DFA(NFA nfa, Collection roots) {
        this.nfa = nfa;
        this.roots = new ArrayList(roots);
        this.dfaStates = null;
        this.nodes = null;
        this.edges = null;
        this.startStates = null;
        this.acceptStates = null;
    }

    public NFAState start() {
        lazyBuild();
        return startStates[nfa.start().index()];
    }

    /**
     * We avoid building the DFA until it's used.
     * All methods in the public interface must call lazyBuild before
     * they do their work.
     */
    protected void lazyBuild() {
        if (this.nodes == null) {
System.out.println("building DFA for " + nfa);
            this.dfaStates = new HashMap();
            subsetConstruction(roots);
            this.nodes = new ArrayList(dfaStates.values());
            this.dfaStates = null; // free up some memory
            dump();
        }
    }

    public Iterator nodes() {
        lazyBuild();
        return nodes.iterator();
    }

    public NFAState nodeAt(int i) {
        lazyBuild();
        return (NFAState) nodes.get(i);
    }

    public List alphabet() {
        return nfa.alphabet();
    }

    public int size() {
        lazyBuild();
        return nodes.size();
    }

    public Iterator succs(Node n) {
        lazyBuild();
        Collection out = (Collection) edges.get(n);
        if (out == null) {
            return Collections.EMPTY_LIST.iterator();
        }
        return out.iterator();
    }

    public Iterator preds(Node n) {
        throw new InternalCompilerError("unimplemented");
    }

    public void dump() {
        if (Report.should_report(TOPICS, 3)) {
            Report.report(1, "Dumping " + this);

            if (this.nodes == null) {
                System.err.println("DFA not built yet");
                return;
            }

            for (int i = 0; i < startStates.length; i++) {
                if (startStates[i] == null) continue;
                System.err.println("start state " + startStates[i]);
            }
            for (Iterator i = nodes(); i.hasNext(); ) {
                DFAState T = (DFAState) i.next();
                System.err.println("state " + T);
                for (Iterator j = succs(T); j.hasNext(); ) {
                    Edge e = (Edge) j.next();
                    System.err.println("edge " + e);
                }
            }

            System.err.println();
            System.err.println("digraph G {");

            for (Iterator i = nodes(); i.hasNext(); ) {
                DFAState n = (DFAState) i.next();
                System.err.println("\"" + n + "\"");
                for (Iterator j = succs(n); j.hasNext(); ) {
                    DFAEdge e = (DFAEdge) j.next();
                    System.err.println("/* " + e + " */");
                    System.err.print("  \"" + e.source() + "\"");
                    System.err.print("->\"" + e.target() + "\"");
                    if (e.label() != null) {
                        System.err.println(" [ label=\"" + e.label() + "\" ]");
                    }
                    else {
                        System.err.println();
                    }
                }
            }

            System.err.println("}");
            System.err.println();
        }
    }

    /**
     * Return true if the language represented by this DFA intersect the
     * language of <code>dfa2</code>.
     */
    public boolean intersects(DFA dfa2) {
        DFA dfa1 = this;

        dfa1.lazyBuild();
        dfa2.lazyBuild();

        // Conceptually, we construct the intersection of two DFAs
        // M(1) and M(2) as follows:
        //
        // Take the cross product of the states in M(1) and M(2)
        // (i1,i2) is a state in the intersection DFA.
        // There is an edge from (i1,i2) to (j1,j2) if
        // there is an edge from i1 to j1 in M(1) _and_ from i2 to j2 in M(2).
        // (i1,i2) is an accepting state if i1 is an accepting state in M(1)
        // and i2 is an accepting state in M(2).

        // But, we just need to check if the intersection accepts
        // _any_ string.  So, just do a depth-first-search on both DFAs
        // at once, constructing the intersection DFA on-the-fly.

        for (int i = 0; i < dfa1.startStates.length; i++) {
            if (dfa1.startStates[i] == null) continue;

            for (int j = 0; j < dfa2.startStates.length; j++) {
                if (dfa2.startStates[j] == null) continue;

                boolean result = intersectDFS(dfa1.startStates[i],
                                              dfa2.startStates[j],
                                              dfa1.startStates[i],
                                              dfa2.startStates[j],
                                              dfa1, dfa2, new BitSet());
                if (result) {
                    return true;
                }
            }
        }

        return false;
    }

    private static int indent = 0;

    private static boolean intersectDFS(DFAState T1, DFAState T2,
                                DFAState start1, DFAState start2,
                                DFA dfa1, DFA dfa2, BitSet visited)
    {
        if (Report.should_report("dfa", 4)) {
            Report.report(4+indent, "intersectDFS: " + T1 + " & " + T2);
        }

        if (dfa1.accepting(start1, T1) && dfa2.accepting(start2, T2)) {
            if (Report.should_report("dfa", 4)) {
                Report.report(4+indent, "intersectDFS: accept for " + start1 + " & " + start2);
            }
            return true;
        }

        int key = T1.index() * dfa2.size() + T2.index();

        if (visited.get(key)) {
            if (Report.should_report("dfa", 4)) {
                Report.report(4+indent, "intersectDFS: visited");
            }
            return false;
        }

        visited.set(key);

        Map labels1 = new HashMap();
        Map labels2 = new HashMap();

        for (Iterator i = dfa1.succs(T1); i.hasNext(); ) {
            DFAEdge e = (DFAEdge) i.next();
            labels1.put(e.label, e);
        }

        for (Iterator i = dfa2.succs(T2); i.hasNext(); ) {
            DFAEdge e = (DFAEdge) i.next();
            labels2.put(e.label, e);
        }

        // Intersect the edges.
        for (Iterator i = labels1.keySet().iterator(); i.hasNext(); ) {
            NFATerminal a = (NFATerminal) i.next();
            Edge e1 = (Edge) labels1.get(a);
            Edge e2 = (Edge) labels2.get(a);

            if (e1 == null) continue;
            if (e2 == null) continue;

            if (Report.should_report("dfa", 4)) {
                Report.report(4+indent, "intersectDFS: common label " + a);
                indent++;
            }

            // Both DFAs have an edge labeled a.
            boolean result = intersectDFS((DFAState) e1.target(),
                                          (DFAState) e2.target(),
                                          start1, start2,
                                          dfa1, dfa2, visited);

            if (Report.should_report("dfa", 4)) {
                indent--;
                Report.report(4+indent, "intersectDFS: call for " + a + " = " + result);
                Report.report(4+indent, "intersectDFS: back at " + T1 + " & " + T2);
            }

            if (result) {
                if (Report.should_report("dfa", 4)) {
                    Report.report(4+indent, "intersectDFS: returning true");
                }
                return true;
            }
        }

        if (Report.should_report("dfa", 4)) {
            Report.report(4+indent, "intersectDFS: returning false");
        }

        return false;
    }

    /** Does the DFA accept in T for the language defined by the start state. */
    public boolean accepting(NFAState q0, NFAState q1) {
        return dfaAccepting((DFAState) q0, (DFAState) q1);
    }

    private boolean dfaAccepting(DFAState start, DFAState T) {
        lazyBuild();

        for (int i = start.nfaStates.nextSetBit(0); i >= 0;
             i = start.nfaStates.nextSetBit(i+1)) {

            NFAState q0 = nfa.nodeAt(i);

            for (int j = T.nfaStates.nextSetBit(0); j >= 0;
                j = T.nfaStates.nextSetBit(j+1)) {

                NFAState q = nfa.nodeAt(j);

                if (nfa.accepting(q0, q)) {
                    return true;
                }
            }
        }

        return false;
    }

    /** Builds a DFA from an NFA.  The algorithm is from the Dragon book. */
    private void subsetConstruction(Collection roots) {
        dfaStates = new HashMap();
        LinkedList worklist = new LinkedList();

        edges = new HashMap();
        startStates = new DFAState[nfa.size()];

        // build the closure for the start state, adding it to the
        // worklist
        for (Iterator i = roots.iterator(); i.hasNext(); ) {
            NFAState start = (NFAState) i.next();
            int root = start.index();
            BitSet nfaStates = new BitSet();
            nfaStates.set(start.index());
            DFAState T = getDFAState(nfaStates, worklist);
            startStates[start.index()] = T;
        }

        // for each DFA state in the worklist,
        // compute the transitions out of the state.
        while (! worklist.isEmpty()) {
            DFAState T = (DFAState) worklist.removeFirst();

            for (int i = 0; i < nfa.alphabet().size(); i++) {
                NFATerminal a = (NFATerminal) nfa.alphabet().get(i);
                DFAState U = getDFAState(move(T, a), worklist);

                if (U != null) {
                    Collection c = (Collection) edges.get(T);
                    if (c == null) {
                        c = new HashSet();
                        edges.put(T, c);
                    }
                    c.add(new DFAEdge(T, U, a));
                }
            }
        }

        if (Report.should_report(TOPICS, 3)) {
            Report.report(3, "NFA has " + nfa.size() + " states");
            Report.report(3, "DFA has " + dfaStates.size() + " states");
        }
    }

    private BitSet move(DFAState T, NFATerminal a) {
        BitSet move = new BitSet();

        for (int state = T.nfaStates.nextSetBit(0); state >= 0;
             state = T.nfaStates.nextSetBit(state+1)) {

            for (Iterator i = nfa.succs(nfa.nodeAt(state)); i.hasNext(); ) {
                NFAEdge e = (NFAEdge) i.next();
                if (e.label() != null && e.label().index() == a.index()) {
                    move.set(e.target().index());
                }
            }
        }

        return move;
    }

    private BitSet epsilonClosure(BitSet nfaStates) {
        int[] stack = new int[nfa.size()];
        int top = 0;

        if (Report.should_report(TOPICS, 5)) {
            System.err.print("     e-closure of " + nfaStates);
        }

        for (int state = nfaStates.nextSetBit(0); state >= 0;
             state = nfaStates.nextSetBit(state+1)) {
            stack[top++] = state;
        }

        while (top > 0) {
            int t = stack[--top];

            for (Iterator i = nfa.succs(nfa.nodeAt(t)); i.hasNext(); ) {
                NFAEdge e = (NFAEdge) i.next();
                if (e.label() == null) {
                    if (! nfaStates.get(e.target().index())) {
                        stack[top++] = e.target().index();
                        nfaStates.set(e.target().index());
                    }
                }
            }
        }

        if (Report.should_report(TOPICS, 5)) {
            System.err.println(" is " + nfaStates);
        }

        return nfaStates;
    }

    private DFAState getDFAState(BitSet nfaStates, LinkedList worklist) {
        nfaStates = epsilonClosure(nfaStates);

        if (nfaStates.isEmpty()) {
            return null;
        }

        DFAState T = (DFAState) dfaStates.get(nfaStates);

        if (T == null) {
            T = new DFAState(nfaStates, dfaStates.size());
            dfaStates.put(nfaStates, T);
            worklist.add(T);
        }

        return T;
    }

    protected static class DFAState extends NFAState {
        BitSet nfaStates;

        DFAState(BitSet nfaStates, int index) {
            super(index);
            this.nfaStates = new BitSet();
            this.nfaStates.or(nfaStates);
        }

        public boolean equals(Object o) {
            return o instanceof DFAState && ((DFAState) o).index == index;
        }

        public int hashCode() {
            return index;
        }

        public String toString() {
            // return "<" + index + ": " + nfaStates + ">";
            return "" + index;
        }

        public int index() {
            return index;
        }
    }

    protected static class DFAEdge extends NFAEdge {
        DFAEdge(DFAState source, DFAState target, NFATerminal label) {
            super(source, target, label);
        }
    }

    public String toString() {
        return "DFA(" + nfa + ")";
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "lr", "ibex", "dfa" });
}
