// This file is derived from the BLOAT bytecode optimizer, written
// by Nathaniel Nystrom, Antony Hosking, David Whitlock at Purdue.
// Used by permission of the first author (who wrote this file, too).

package polyglot.ext.ibex.lr;

import polyglot.util.*;
import java.util.*;
import polyglot.main.Report;

/**
 * Dominators finds the dominators of a RootedGraph
 * <p>
 * The algorithm used is Purdum-Moore.  It isn't as fast as Lengauer-Tarjan,
 * but it's a lot simpler.
 */
public class Dominators
{
    // Bit matrix indicating the dominators of each node in the graph.
    // If bit j of dom[i] is set, then node j dominates node i.
    BitSet[] dom;

    RootedGraph graph;

    /**
     * Calculates which nodes in a graph dominate which other nodes.
     *
     * @param graph
     *        The graph that is used to find the dominator tree.
     */  
    public Dominators(RootedGraph graph)
    {
        this.graph = graph;

        int size = graph.size();          // The number of vertices in the cfg
        
        // Get the index of the root
        Node root = graph.root();

        // Bit matrix indicating the dominators of each vertex.
        // If bit j of dom[i] is set, then node j dominates node i.
        this.dom = new BitSet[size];

        // A bit vector of all 1's
        BitSet ALL = new BitSet(size);          

        for (int i = 0; i < size; i++) {
            ALL.set(i);
        }

        // Initially, all the bits in the dominance matrix are set, except
        // for the root node.  The root node is initialized to have itself
        // as an immediate dominator.
        for (int i = 0; i < size; i++) {
            BitSet domI = new BitSet(size);
            domI.or(ALL);
            dom[i] = domI;
        }

        dom[root.index()].clear();
        dom[root.index()].set(root.index());

        // Did the dominator bit vector array change?
        Node[] stack = new Node[size];
        int top = 0;
        BitSet instack = new BitSet();

        // push the successors of root
        for (Iterator k = graph.succs(root); k.hasNext(); ) {
            Node j = ((Edge) k.next()).target();
            if (! instack.get(j.index())) {
                stack[top++] = j;
                instack.set(j.index());
            }
        }

        // Compute the dominators of each node in the graph.  We iterate
        // over every reachable node in the graph.  The dominators of a
        // node, x, are found by taking the intersection of the
        // dominator bit vectors of each predecessor of x and unioning
        // that with x.  This process is repeated until no changes are
        // made to any dominator bit vector.
        // 
        // We should iterate in toposort order to be most efficient;
        // but any order will work.
        while (top > 0) {
            Node i = stack[--top];
            instack.clear(i.index());

            // We already know the dominators of the root.
            if (i.index() == root.index()) {
                continue;
            }

            BitSet oldSet = dom[i.index()];
            BitSet domI = new BitSet(size);
            domI.or(oldSet);

            // domI := intersection of dom(pred) for all pred(i).

            // Find the intersection of the dominators of i's 
            // predecessors.
            for (Iterator k = graph.preds(i); k.hasNext(); ) {
                Node j = ((Edge) k.next()).target();
                if (Report.should_report(TOPICS, 2))
                    Report.report(2, "processing " + j + "->" + i);
                domI.and(dom[j.index()]);
            }

            // Include yourself in your dominators.
            domI.set(i.index());

            if (Report.should_report(TOPICS, 2)) {
                Report.report(2, " old dom(" + i + ") = " + oldSet);
                Report.report(2, " new dom(" + i + ") = " + domI);
            }

            // If the set changed, push the successors.
            if (! domI.equals(oldSet)) {
                for (Iterator k = graph.succs(i); k.hasNext(); ) {
                    Node j = ((Edge) k.next()).target();
                    if (! instack.get(j.index())) {
                        stack[top++] = j;
                        instack.set(j.index());
                    }
                }
                dom[i.index()] = domI;
            }
        }
    }

    /** Return true if <code>j</code> dominates <code>i</code>. */
    public boolean dominates(Node j, Node i) {
        // If bit j of dom[i] is set, then node j dominates node i.
        return dom[i.index()].get(j.index());
    }

    /** Return the set of all dominators of <code>i</code>. */
    public BitSet allDominators(Node i) {
        BitSet x = new BitSet();
        x.or(dom[i.index()]);
        return x;
    }

    /**
     * Return the set of nodes in the <i>dominance frontier</i> of
     * node <code>i</code>.
     * <p>
     * The <i>dominance frontier</i> of a node x is the set of all nodes
     * w such that x dominates a predecessor of w, but does not strictly
     * dominate w.  Essentially, nodes in the dominance frontier have one
     * parent that <b>is</b> dominated by x and at least one parent that
     * <b>is not</b> dominated by x.
     * </p>
     */
    public BitSet DF(Node i) {
        int size = graph.size();

        BitSet DF = new BitSet();

        if (Report.should_report(TOPICS, 2))
            Report.report(2, "computing DF(" + i + ")");

        // Add to DF if i dominates some, but not all predecessors of j.
        for (Iterator j = graph.nodes(); j.hasNext(); ) {
            Node n = (Node) j.next();

            int numNo = 0;
            int numYes = 0;

            for (Iterator k = graph.preds(n); k.hasNext(); ) {
                Node p = ((Edge) k.next()).target();
                if (dominates(i, p)) {
                    if (Report.should_report(TOPICS, 3))
                        Report.report(3, i + " dominates " + p + "->" + n);
                    numYes++;
                }
                else {
                    if (Report.should_report(TOPICS, 3))
                        Report.report(3, i + " does not dominate " + p + "->" + n);
                    numNo++;
                }
            }

            if (numYes > 0 && numNo > 0) {
                if (Report.should_report(TOPICS, 2))
                    Report.report(2, "DF(" + i + ") includes " + n);
                DF.set(n.index());
            }
        }

        return DF;
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "dominators" });
}
