package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class constructs a DFA from the SR graph.
 *
 * Let L(i) be the language accepted by state i.
 *
 * First build an NFA from the SR graph as follows:
 * There is a transition for each edge in the SR graph
 * where S edges are labeled with their token, and
 *       R edges are labeled \epsilon
 * Node i is the accept state for L(i).
 * Node ROOT is the start state.
 *
 * Then perform the subset construction to create the DFA.
 *
 * The DFA is built to answer only one question: is the intersection
 * of two languages L(i) and L(j) empty.  This can be answered simply
 * by checking if there is a state reachable from the root
 * that accepts both L(i) and L(j).
 */
class SRDFA extends DFA
{
    Map edges;
    Map intersectCache;

    SRDFA(SRGraph graph, Collection roots) {
        super(new SRNFA(graph), roots);
        this.intersectCache = new HashMap();
    }

    static class Key {
        DFAState dfaState;
        Node nfaAccept1, nfaAccept2;

        Key(DFAState dfaState, Node nfaAccept1, Node nfaAccept2) {
            this.dfaState = dfaState;
            this.nfaAccept1 = nfaAccept1;
            this.nfaAccept2 = nfaAccept2;
        }

        public int hashCode() {
            return dfaState.hashCode() +
                   nfaAccept1.hashCode() + nfaAccept2.hashCode();
        }

        public boolean equals(Object o) {
            if (o instanceof Key) {
                Key k = (Key) o;
                return dfaState.equals(k.dfaState) &&
                       (nfaAccept1.equals(k.nfaAccept1) &&
                        nfaAccept2.equals(k.nfaAccept2))
                    || (nfaAccept1.equals(k.nfaAccept2) &&
                        nfaAccept2.equals(k.nfaAccept1));
            }
            return false;
        }
    }

    /**
     * Is the intersection of the two languages L_root(acceptState1) and
     * L_root(acceptState2) non-empty?
     */
    boolean intersectionNonEmpty(SRNode root,
        SRNode acceptState1, SRNode acceptState2)
    {
        lazyBuild();

        // The usual construction for the intersection of two DFAs
        // M(1) and M(2) is.
        // 
        // Take the cross product of the states in M(1) and M(2)
        // (i1,i2) is a state in the intersection DFA.
        // There is an edge from (i1,i2) to (j1,j2) if 
        // there is an edge from i1 to j1 in M(1) _and_ from i2 to j2 in M(2).
        // (i1,i2) is an accepting state if i1 is an accepting state in M(1)
        // and i2 is an accepting state in M(2).
        // 
        // Now, since the M(1) and M(2) both share the same edges,

        if (startStates[root.index()] == null) {
            throw new InternalCompilerError(root + " is not an NFA start state.");
        }

        return nonEmptyDFS(startStates[root.index()],
                           acceptState1, acceptState2, new BitSet());
    }

    boolean nonEmptyDFS(DFAState dfaState, Node nfaAccept1, Node nfaAccept2,
                        BitSet visited)
    {
        Key key = new Key(dfaState, nfaAccept1, nfaAccept2);
        Boolean b = (Boolean) intersectCache.get(key);

        if (b != null) {
            return b.booleanValue();
        }

        if (visited.get(dfaState.index())) {
            intersectCache.put(key, Boolean.FALSE);
            return false;
        }

        visited.set(dfaState.index());

        if (dfaState.nfaStates.get(nfaAccept1.index()) &&
            dfaState.nfaStates.get(nfaAccept2.index())) {
            intersectCache.put(key, Boolean.TRUE);
            return true;
        }

        for (Iterator i = succs(dfaState); i.hasNext(); ) {
            DFAEdge e = (DFAEdge) i.next();
            if (nonEmptyDFS((DFAState) e.target(),
                            nfaAccept1, nfaAccept2, visited)) {
                intersectCache.put(key, Boolean.TRUE);
                return true;
            }
        }

        intersectCache.put(key, Boolean.FALSE);
        return false;
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "lr", "ibex", "dfa" });
}
