package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;

class SRGraph implements Graph {
    Map fwdGraph;
    Map revGraph;
    BitSet accepting;
    Set[] statesMap;
    List startStates;
    List nodes;
    Map nodeCache;
    List terminals;

    SRGraph(LRConstruction lrc) {
        this.fwdGraph = new HashMap();
        this.revGraph = new HashMap();
        this.accepting = new BitSet();
        this.statesMap = new Set[lrc.states.size()];
        this.nodes = new ArrayList();
        this.nodeCache = new HashMap();
        this.terminals = lrc.terminals;

        for (Iterator i = lrc.startStates.iterator(); i.hasNext(); ) {
            State I = (State) i.next();
            SRNode n = getNode(I, null);
            statesMap[I.index()] = new HashSet();
            statesMap[I.index()].add(n);
        }

        Set reduceVisited = new HashSet();

        for (int i = 0; i < lrc.states.size(); i++) {
            State I = (State) lrc.states.get(i);

            Set[] actionRow = lrc.actionTable[i];

            for (int j = 0; j < actionRow.length; j++) {
                for (Iterator k = actionRow[j].iterator(); k.hasNext(); ) {
                    Action a = (Action) k.next();
                    if (a instanceof Reduce) {
                        Reduce r = (Reduce) a;
                        addReduceEdges(lrc, I, r.rule, reduceVisited);
                    }
                    else if (a instanceof Shift) {
                        Shift s = (Shift) a;
                        addShiftEdges(lrc, I, s);
                    }
                    else if (a instanceof Accept) {
                        setAccept(I);
                    }
                }
            }
        }

        dump();
    }

    public Iterator nodes() {
        return nodes.iterator();
    }

    Collection nodesForParserStates(Collection states) {
        Set s = new HashSet();

        for (Iterator i = states.iterator(); i.hasNext(); ) {
            State I = (State) i.next();

            for (Iterator nodes = nodesForParserState(I).iterator();
                nodes.hasNext(); ) {

                SRNode n = (SRNode) nodes.next();
                s.add(n);
            }
        }

        return s;
    }

    Collection nodesForParserState(State I) {
        if (statesMap[I.index()] == null) {
            return Collections.EMPTY_LIST;
        }
        else {
            return statesMap[I.index()];
        }
    }

    static class DFSKey {
        State I; State J; State top; GLRRule rule; int depth;

        DFSKey(State I, State J, GLRRule rule, int depth, State top) {
            this.I = I;
            this.J = J;
            this.rule = rule;
            this.depth = depth;
            this.top = top;
        }

        public int hashCode() {
            return I.hashCode() + J.hashCode() + rule.hashCode() + depth;
        }

        public boolean equals(Object o) {
            if (o instanceof DFSKey) {
                DFSKey k = (DFSKey) o;
                return I.equals(k.I) && J.equals(k.J) && rule.equals(k.rule) && depth == k.depth && (top == null ? k.top == null : top.equals(k.top));
            }
            return false;
        }
    }

    // Add edges from I to any state that can be transitioned to by
    // reducing using rule.
    void addReduceEdges(LRConstruction lrc, State I, GLRRule rule, Set visited) {
        addReduceEdgesDFS(lrc, I, I, rule, rule.rhs.size(), null, visited);
    }

    void addReduceEdgesDFS(LRConstruction lrc, State I, State J, GLRRule rule, int depth, State top, Set visited) {
        DFSKey dfsKey = new DFSKey(I, J, rule, depth, top);

        if (visited.contains(dfsKey)) {
            return;
        }

        visited.add(dfsKey);

        if (depth == 0) {
            List tops;
            if (top != null) {
                tops = Collections.singletonList(top);
            }
            else {
                // The rule was A -> \epsilon
                // We're reducing it in state I = J, 
                // The top of stack could be any state that precedes J.
                tops = allPredStates(lrc, J);

                if (tops.isEmpty()) {
                    tops.add(null);
                }
            }

            // create an edge from <I,top> to <goto(J,LHS),J>.
            State K = lrc.gotoState(J, rule.lhs);

            for (Iterator i = tops.iterator(); i.hasNext(); ) {
                State H = (State) i.next();
                addReduceEdge(I, H, rule, K, J);
            }
        }
        else {
            // Find all edges into J labeled with the symbol we've popped to.
            GLRSymbol sym = (GLRSymbol) rule.rhs.get(depth-1);

            for (Iterator k = lrc.reverseGotoOrShiftStates(J, sym).iterator();
                 k.hasNext(); ) {

                // There is an edge from K to J labeled with sym.
                State K = (State) k.next();

                if (depth == rule.rhs.size()) {
                    top = K;
                }

                addReduceEdgesDFS(lrc, I, K, rule, depth-1, top, visited);
            }
        }
    }

    List allPredStates(LRConstruction lrc, State J) {
        List preds = new ArrayList();

        for (Iterator i = lrc.nonterminals.iterator(); i.hasNext(); ) {
            GLRSymbol sym = (GLRSymbol) i.next();
            for (Iterator k = lrc.reverseGotoOrShiftStates(J, sym).iterator();
                k.hasNext(); ) {
                // There is an edge from K to J labeled with sym.
                State K = (State) k.next();
                preds.add(K);
            }
        }

        for (Iterator i = lrc.terminals.iterator(); i.hasNext(); ) {
            GLRSymbol sym = (GLRSymbol) i.next();
            for (Iterator k = lrc.reverseGotoOrShiftStates(J, sym).iterator();
                k.hasNext(); ) {
                // There is an edge from K to J labeled with sym.
                State K = (State) k.next();
                preds.add(K);
            }
        }

        return preds;
    }

    public void dump() {
        if (Report.should_report(TOPICS, 3)) {
            Report.report(1, "Dumping SR graph:");

            System.out.println("digraph G {");

            for (Iterator i = nodes.iterator(); i.hasNext(); ) {
                SRNode n = (SRNode) i.next();
                System.out.println("\"" + n + "\"");
                if (accepting(n.state)) {
                    System.out.println(" /* " + n + " accepts */");
                }
                for (Iterator j = succs(n); j.hasNext(); ) {
                    Edge e = (Edge) j.next();
                    System.out.println("/* " + e + " */");
                    System.out.print("  \"" + e.source() + "\"");
                    System.out.print("->\"" + e.target() + "\"");
                    if (e instanceof ShiftEdge) {
                        System.out.println(" [ label=\"" + ((ShiftEdge) e).label() + "\" ]");
                    }
                    else if (e instanceof ReduceEdge) {
                        System.out.println(" [ label=\"" + ((ReduceEdge) e).reduction + "\" ]");
                    }
                }
            }

            System.out.println("}");
        }
    }

    void setAccept(State I) {
        accepting.set(I.index());
    }

    boolean accepting(State I) {
        return accepting.get(I.index());
    }

    // There is a shift(x) edge from <s1,top1> to <s2,s1>
    // if there is a shift of x in s1 to s2.
    void addShiftEdges(LRConstruction lrc, State J, Shift s) {
        // Get all the states K that could be on the top of stack below J
        // and add shift edges from <J,K> to <s.dest,J>.
        List preds = allPredStates(lrc, J);

        if (preds.isEmpty()) {
            preds.add(null);
        }

        for (Iterator k = preds.iterator(); k.hasNext(); ) {
            State K = (State) k.next();

            // There is an edge from K to J labeled with sym.
            SRNode src = getNode(J, K);
            SRNode dst = getNode(s.dest, J);

            SREdge e = new ShiftEdge(src, dst, s.terminal);
            addEdge(src, dst, e);

            if (Report.should_report(TOPICS, 3))
                Report.report(3, "shift edge from " + src + " to " + dst);
        }
    }

    // Add a shift edge from <I,top> to <K,J>.
    void addReduceEdge(State I, State top, GLRRule rule, State K, State J) {
        SRNode src = getNode(I, top);
        SRNode dst = getNode(K, J);

        SREdge e = new ReduceEdge(src, dst, rule);
        addEdge(src, dst, e);

        if (Report.should_report(TOPICS, 3))
            Report.report(3, "reduce edge from " + I + " to " + K + " for " + rule + " with top of stack " + J);
    }

    public int size() {
        return nodes.size();
    }

    public Iterator succs(Node n) {
        Collection c = (Collection) fwdGraph.get(n);
        if (c == null) return Collections.EMPTY_LIST.iterator();
        return c.iterator();
    }

    public Iterator preds(Node n) {
        Collection c = (Collection) revGraph.get(n);
        if (c == null) return Collections.EMPTY_LIST.iterator();
        return c.iterator();
    }

    public SRNode nodeAt(int k) {
        return (SRNode) nodes.get(k);
    }

    SRNode getNode(State state, State top) {
        SRNode key = new SRNode(state, top, 0);
        SRNode n = (SRNode) nodeCache.get(key);
        if (n == null) {
            n = new SRNode(state, top, nodes.size());
            nodes.add(n);
            nodeCache.put(key, n);

            if (statesMap[state.index()] == null) {
                statesMap[state.index()] = new HashSet();
            }
            statesMap[state.index()].add(n);
        }
        return n;
    }

    void addEdge(SRNode src, SRNode dst, Edge e) {
        Collection c;

        c = (Collection) fwdGraph.get(src);
        if (c == null) {
            c = new HashSet();
            fwdGraph.put(src, c);
        }
        c.add(e);

        c = (Collection) revGraph.get(dst);
        if (c == null) {
            c = new HashSet();
            revGraph.put(dst, c);
        }
        c.add(new RevEdge(e));
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "lr", "ibex", "srgraph" });
}
