package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class contains code that converts a context free grammar into
 * a DFA for a regular approximation of the language.
 *
 * The basic translation is due to Nederhof and is decribed in the paper
 * "Practical Experiments with Regular Approximation" in Computational
 * Linguistics, 26(1).  Section 4.1.
 */
class SelfRecursiveApproximationFA extends AbstractNFA
{
    /**
     * The history is the set of all items in the stack below
     * this point.  This can be HUGE.
     */
    static final int ITEMSET = 1;

    /**
     * The history is the set of all items in the stack below
     * this point.  This can be HUGE.
     */
    static final int STACK = 0;

    /**
     * Like item set, but just considers the rule of each item.
     */
    static final int RULESET = 3;

    static final int HISTORY_KIND = STACK;
    static final int HISTORY_LEN = 1;

    Map ntStates;
    List allStates;
    NFAState[] startStates;
    NFAState[] acceptStates;
    ApproximationFA afa;
    Collection startSymbols;

    /**
     * Build a regular approximation for the rule's RHS.
     * The LHS must be a augmented grammar start symbol; it cannot
     * be reachable from the RHS.
     */
    SelfRecursiveApproximationFA(List terminals, List nonterminals,
                                 ApproximationFA afa, Collection roots) {
        super(terminals);
        this.afa = afa;
        this.startSymbols = roots;
        this.startStates = new NFAState[nonterminals.size()];
        this.acceptStates = new NFAState[nonterminals.size()];
        this.allStates = new ArrayList();
        this.ntStates = new HashMap();

        this.terminals = new ArrayList(terminals);

        for (Iterator i = startSymbols.iterator(); i.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) i.next();

            History newHistory = newHistory();
            buildNFA(A, newHistory);

            this.startStates[A.index()]  = nonterminalState(A, 0, newHistory);
            this.acceptStates[A.index()] = nonterminalState(A, 1, newHistory);
        }

    }

    History newHistory() {
        if (HISTORY_KIND == ITEMSET) {
            return new ItemSet();
        }
        else if (HISTORY_KIND == RULESET) {
            return new RuleSet();
        }
        else {
            return new Stack();
        }
    }

    public String toString() {
        return "self-recursive approximation NFA for " + startSymbols;
    }

    public NFAEdge edge(NFAState q1, NFATerminal label, NFAState q2) {
        return new NFAEdge(q1, q2, label);
    }

    public NFAState nodeAt(int i) {
        return (NFAState) allStates.get(i);
    }

    public Iterator nodes() {
        return allStates.iterator();
    }

    public List startStates() {
        List l = new ArrayList();
        for (int i = 0; i < startStates.length; i++) {
            if (startStates[i] != null) {
                l.add(startStates[i]);
            }
        }
        return l;
    }

    public NFAState start() { throw new InternalCompilerError("bad NFA"); }
    public NFAState accept() { throw new InternalCompilerError("bad NFA"); }

    public boolean accepting(NFAState q0, NFAState q) {
        for (int i = 0; i < startStates.length; i++) {
            if (q0 == startStates[i] && q == acceptStates[i]) {
                return true;
            }
        }
        return false;
    }

    public int size() { return allStates.size(); }

    static class Key {
        GLRNonterminal A;
        int i;
        History history;

        Key(GLRNonterminal A, int i, History history) {
            this.A = A;
            this.i = i;
            this.history = history;
        }

        public String toString() {
            return A + " " + i + " " + history;
        }

        public int hashCode() { return A.hashCode() + i; }

        public boolean equals(Object o) {
            if (o instanceof Key) {
                Key k = (Key) o;
                return A.equals(k.A) && i == k.i && history.equals(k.history);
            }
            return false;
        }
    }

    NFAState nonterminalState(GLRNonterminal A, int i, History history) {
        return (NFAState) ntStates.get(new Key(A, i, history));
    }

    NFAState createNonterminalState(GLRNonterminal A, int i, History history) {
        NFAState q = new NonterminalState(A, i, history, allStates.size());
        allStates.add(q);
        ntStates.put(new Key(A, i, history), q);
        return q;
    }

    static class HistoryState extends NFAState {
        History history;
        
        HistoryState(History history, int index) {
            super(index);
            this.history = history;
        }
    }

    static class NonterminalState extends HistoryState {
        GLRNonterminal A;
        int i;

        NonterminalState(GLRNonterminal A, int i, History history, int index) {
            super(history, index);
            this.A = A;
            this.i = i;
        }

        public String toString() {
            return A + "(" + i + "," + history + ")";
        }
    }

    static class ItemState extends HistoryState {
        Item item;
        ItemState(Item item, History history, int index) {
            super(history, index);
            this.item = item;
        }
        public String toString() {
            return "item(" + item + "," + history + ")";
        }
    }

    NFAState createItemState(Item item, History history) {
        NFAState q = new ItemState(item, history, allStates.size());
        allStates.add(q);
        return q;
    }

    /**
     * Build and NFA for nonterminal A with the given stack.
     * Invariant: the stack should be truncated before the call.
     * This prevents the truncation from being recomputed multiple
     * times.
     */
    void buildNFA(GLRNonterminal A, History history) {
        if (Report.should_report(TOPICS, 4)) {
            Report.report(4, "buildNFA: nonterminal " + A);
            Report.report(4, "buildNFA: history " + history);
        }

        if (nonterminalState(A, 0, history) != null) {
            // already been here
            return;
        }

        int Ni = afa.partitions.find(A.index());

        // Create two states for each nonterminal
        NFAState qA1 = createNonterminalState(A, 0, history);
        NFAState qA2 = createNonterminalState(A, 1, history);

        // Create a state for each element of the RHS,
        // threading the states together.
        for (Iterator i = A.rules().iterator(); i.hasNext(); ) {
            GLRRule rule = (GLRRule) i.next();

            if (Report.should_report(TOPICS, 4)) {
                Report.report(4, "buildNFA: rule " + rule);
            }

            Item item = new Item(rule, 0, Collections.EMPTY_SET);

            NFAState q0 = createItemState(item, history);

            epsilon(qA1, q0);

            NFAState qprev = q0;

            while (item.afterDot() != null) {
                GLRSymbol X = (GLRSymbol) item.afterDot();
                Item prev = item;

                item = item.shiftDot();

                if (Report.should_report(TOPICS, 4)) {
                    Report.report(4, "buildNFA: symbol " + X);
                    Report.report(4, "buildNFA: item " + item);
                }

                NFAState qj = createItemState(item, history);

                if (X instanceof GLRNonterminal) {
                    GLRNonterminal B = (GLRNonterminal) X;

                    int Nj = afa.partitions.find(B.index());

                    if (Ni == Nj || afa.recursive(Nj) == ApproximationFA.NONE) {
                        History newHistory = history.push(prev);

                        buildNFA(B, newHistory);

                        NFAState qB1 = nonterminalState(B, 0, newHistory);
                        NFAState qB2 = nonterminalState(B, 1, newHistory);

                        epsilon(qprev, qB1);
                        epsilon(qB2, qj);
                    }
                    else {
                        transition(qprev, afa.ntTerminal(B), qj);
                    }
                }
                else {
                    transition(qprev, (GLRTerminal) X, qj);
                }

                qprev = qj;
            }

            epsilon(qprev, qA2);
        }
    }

    static abstract class History {
        abstract History push(Item item);
        public abstract int hashCode();
        public abstract boolean equals(Object o);
    }

    static class ItemSet extends History {
        Set set;

        ItemSet() {
            set = Collections.EMPTY_SET;
        }

        History push(Item item) {
            ItemSet s = new ItemSet();
            s.set = new HashSet(this.set);
            s.set.add(item);
            return s;
        }

        public int hashCode() {
            return set.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof ItemSet && set.equals(((ItemSet) o).set);
        }

        public String toString() {
            return set.toString();
        }
    }

    static class RuleSet extends History {
        Set set;

        RuleSet() {
            set = Collections.EMPTY_SET;
        }

        History push(Item item) {
            RuleSet s = new RuleSet();
            s.set = new HashSet(this.set);
            s.set.add(item.rule);
            return s;
        }

        public int hashCode() {
            return set.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof RuleSet && set.equals(((RuleSet) o).set);
        }

        public String toString() {
            return set.toString();
        }
    }

    static class Hybrid extends History {
        History s;
        History r;

        Hybrid(History s, History r) {
            this.s = s;
            this.r = r;
        }

        History push(Item item) {
            return new Hybrid(s.push(item), r.push(item));
        }

        public int hashCode() {
            return s.hashCode() + r.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof Hybrid
                && s.equals(((Hybrid) o).s)
                && r.equals(((Hybrid) o).r);
        }

        public String toString() {
            return s.toString() + " & " + r.toString();
        }
    }

    static class Stack extends History {
        List stack;

        Stack() {
            stack = Collections.EMPTY_LIST;
        }

        History push(Item item) {
            Stack s = new Stack();
            if (HISTORY_LEN == 1) {
                s.stack = new ArrayList(this.stack.size());
                s.stack.add(item);
            }
            else if (HISTORY_LEN > 1) {
                if (this.stack.size() == HISTORY_LEN) {
                    s.stack = new ArrayList(this.stack.size());
                    s.stack.addAll(this.stack.subList(1,this.stack.size()-1));
                    s.stack.add(item);
                }
                else {
                    s.stack = new ArrayList(this.stack.size()+1);
                    s.stack.addAll(this.stack);
                    s.stack.add(item);
                }
            }
            return s;
        }

        public int hashCode() {
            return stack.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof Stack && stack.equals(((Stack) o).stack);
        }

        public String toString() {
            return stack.toString();
        }
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "lr", "ibex", "nfa" });
}
