package ibex.runtime;

import java.io.*;
import java.util.*;

public class GLRParser {
    ParserImpl parser;

    public GLRParser(ParserImpl parser) {
        this.parser = parser;
        actionTable = decodeActionTable();
        overflowTable = decodeOverflowTable();
        gotoTable = decodeGotoTable();
        ruleTable = decodeRuleTable();
    }

    /** Tag used in the action table to indicate an error. */
    public static final int ERROR = 0;

    /**
     * Tag used in the action table to indicate a shift action.
     * The entry also contains the destination state of the shift.
     */
    public static final int SHIFT = 1;

    /**
     * Tag used in the action table to indicate a reduce action.
     * The entry also contains the index of the rule to reduce.
     */
    public static final int REDUCE = 2;

    /**
     * Tag used in the action table to indicate an accept action.
     * The entry also contains the index of the start symbol to accept.
     */
    public static final int ACCEPT = 3;

    /**
     * Tag used in the action table to indicate a conflict.
     * The entry also contains the index into the overflowTable.
     * of the list of conflicting actions.
     */
    public static final int OVERFLOW = 4;

    /**
     * The LR action table.
     * Indexed by parser state and by a terminal index.
     * Entries are either ERROR, SHIFT(state), REDUCE(rule), OVERFLOW(i), 
     * or ACCEPT(symbol).
     */
    int[][] actionTable;

    /**
     * The action overflow table; used when there is a conflict in
     * actionTable.  When an action is OVERFLOW(i), i is used to
     * locate a string of actions in the overflow table:
     * overflowTable[i] = N is the length of the list of actions,
     * overflowTable[i+1 .. i+1+N].
     * Each entry is in the same format as in the actionTable,
     * except the length entries.
     */
    int[] overflowTable;

    /**
     * The goto table, indexed by parser state and a nonterminal.
     */
    int[][] gotoTable;

    /**
     * The rule table, indexed by rule number.
     * The high 24 bits of each entry are the index of the nonterminal.
     * The low 8 bits are the rhs length.
     */
    int[] ruleTable;

    /** The topmost nodes in the GSS. */
    ArrayList topmost;

    /** A node in the graph-structured stack (GSS) */
    protected static class Node {
        /** The state the node represents. */
        int state;

        /** Number of links into the node.  This is used to determine when to
         * recompute the deteministic depths for the entire GSS. */ 
        int refcount;

        /**
         * Deterministic depth of the node.  This is the number of
         * links that can be traversed before reaching a node with
         * outdegree &gt; 1.
         */
        int depth;    

        /* Head of a list of links to the node's children. */
        Link out;     

        public Node(int state) {
            this.state = state;
            this.out = null;
            this.refcount = 0;
            this.depth = 1;
        }

        public String toString() {
            return "" + state;
        }
    }

    protected static class Link {
        Node bottom; // pointer does the stack
        Action semAction; // the semantic action associated with the link
        Link next;   // a sibling link, or null
        int span;

        Link(Node bottom, Node top, Action semAction, int span) {
            this.bottom = bottom;
            this.semAction = semAction;
            this.span = span;
        }

        public String toString() {
            return bottom.toString();
        }
    }

    protected abstract class Action {
        abstract Object run();
    }

    protected class TerminalAction extends Action {
        Terminal terminal;

        TerminalAction(Terminal terminal) {
            this.terminal = terminal;
        }

        Object run() {
            return terminal;
        }

        public String toString() {
            return terminal.toString();
        }
    }

    protected class SemanticAction extends Action {
        int rule;
        Action[] children;
        Object sval;
        SemanticAction ambiguous;

        SemanticAction(int rule, Action[] children) {
            this.rule = rule;
            this.children = children;
        }

        Object run() {
            // Check if already run.
            if (children == null) {
                return sval;
            }

            // Run the children actions, building arguments to pass
            // to the semantic action function.
            // Actions for children should be run in the same order
            // as they appear on the rhs of the rule; that is,
            // svals[0] should be the semantic value for the leftmost
            // symbol on the rhs of the rule.  This is the same order
            // as the children array.
            Object[] svals = new Object[children.length];

            for (int i = 0; i < children.length; i++) {
                svals[i] = children[i].run();
            }

            // Now merge if necessary.
            if (ambiguous != null) {
                // get the other semantic value, recursively running any
                // merges in the tail of the ambiguous action list.
                Object[] svals2 = new Object[ambiguous.children.length];

                for (int i = 0; i < ambiguous.children.length; i++) {
                    svals2[i] = ambiguous.children[i].run();
                }

                // The lower numbered rule should be first.
                if (rule <= ambiguous.rule) {
                    sval = parser.mergeAction(rule, ambiguous.rule, svals, svals2);
                }
                else {
                    sval = parser.mergeAction(ambiguous.rule, rule, svals2, svals);
                }
            }
            else {
                // Apply the semantic action.
                sval = parser.semanticAction(rule, svals);
            }

            // Set children to null to indicate that we've run
            // And to permit the tree to be collected.
            children = null;

            return sval;
        }

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("action(rule #");
            sb.append(rule);
            sb.append(": ");

            for (int i = 0; i < children.length; i++) {
                if (i > 0) sb.append(", ");
                sb.append(children[i].toString());
            }

            sb.append(")");

            return sb.toString();
        }
    }

    public void error(Node n, Terminal t) {
        // pop the stack until a recovery state s is found.
        // state s corresponds to the item A -> error . alpha

        // discard input until a token in FIRST(alpha), or
        // FOLLOW(A) if alpha is empty, is found.
        // Then resume parsing.
        // Recovery is complete when A -> error alpha .
        // is reduced.

        // the semantic value for all reductions performed in recovery
        // is ERROR.  Do not call apply.

        System.out.println("error in state " + n.state + " at " + t);
        System.out.print("expected one of:");
        for (int i = 0; i < actionTable[n.state].length; i++) {
            if (action(actionTable[n.state][i]) != ERROR) {
                System.out.print(" " + i);
            }
        }
        System.out.println();

        error = true;
    }

    boolean error;

    // returns the set of accepting stack nodes
    public Object parse(int startState, int startSym) throws IOException {
        // Pre-allocate the return array of reductions().  The parser can, at
        // most, reduce using all rules at once, plus one to indicate the end.
        globalRules = new int[ruleTable.length+1];

        error = false;

        topmost = new ArrayList();
        topmost.add(new Node(startState));
        boolean accept = false;

        EOF eof = new EOF(parser.eofSymbol());

SCAN:
        while (! accept && ! error) {
            Terminal t;
            
            try {
                t = parser.scan();
                if (t.symbol() == parser.eof().symbol()) {
                    t = eof;
                }
            }
            catch (EOFException e) {
                t = eof;
            }

CHURN:
            while (! accept) {
                System.out.println("------------------------");
                System.out.println("scan: " + t);
                System.out.println("topmost: " + topmost);

                // Do LR(1) parsing if it will be deterministic.
                if (false && topmost.size() == 1) {
                    Node n = (Node) topmost.get(0);
                    int e = actionTable[n.state][t.symbol()];
                    int action = action(e);

                    switch (action) {
                        case ERROR:
                            error(n, t);
                            break SCAN;
                        case SHIFT: {
                            int dest = actionData(e);

                            System.out.println("shift " + t + " to " + dest);

                            Node topSib = new Node(dest);

                            // insert topSib into topmost;
                            // push topSib onto stack
                            addLink(n, topSib, new TerminalAction(t), 1);

                            topmost.remove(0);
                            topmost.add(topSib);
                            continue SCAN;
                        }
                        case REDUCE: {
                            int rule = actionData(e);
                            int lhs = ruleLhsIndex(ruleTable[rule]);
                            int rhsLength = ruleRhsLength(ruleTable[rule]);

                            if (n.depth >= rhsLength) {
                                // the reduce is deterministic

                                Action[] semAction = new Action[rhsLength];

                                // walk the stack, building the array of
                                // children for the semantic action node
                                Node current = n;
                                int span = 0;

                                for (int i = 0; i < rhsLength; i++) {
                                    Link l = current.out;
                                    semAction[i] = l.semAction;
                                    current = l.bottom;
                                    span += l.span;
                                }

                                // create a semantic action node.
                                SemanticAction v = new SemanticAction(rule, semAction);

                                // current now points to the new top of stack

                                int nextState = gotoTable[current.state][lhs];

                            System.out.println("reduce with rule " + rule);
                            System.out.println("        and goto " + nextState);

                                Node newTop = new Node(nextState);

                                // discard
                                if (current != n) current.refcount--;
                                topmost.remove(0);
                                topmost.add(newTop);

                                addLink(current, newTop, v, span);

                                // resume at the next token
                                continue CHURN;
                            }
                            else {
                                // nondeterministic reduce; break to the GLR code
                                break;
                            }
                        }
                        case OVERFLOW:
                              System.out.println("overflow!");
                              // break to the GLR code;
                              break;
                        case ACCEPT:
                              // let the GLR code handle it.
                              break;
                    }
                }

                System.out.println("doing GLR parse");

                System.out.println("before R topmost = " + topmost);

                // Couldn't do LR(1) parsing.  Do GLR parsing.
                doReductions(t);

                System.out.println("before S topmost = " + topmost);

                accept = doAccepts(t, startSym);

                if (! accept) {
                    doShifts(t);
                }

                System.out.println("after S topmost = " + topmost);

                break CHURN;
            }
        }

        if (! error && ! topmost.isEmpty()) {
            // Only nodes in an accepting state should be in topmost.
            // Get all the semantic values in all the accepting states.
            List l = filter(topmost, startSym);

            if (l.size() == 0) {
                System.out.println("No semantic value.");
            }
            else if (l.size() > 1) {
                System.out.println("Unresolved ambiguity: " + l);
            }

            System.out.println("Returning " + l.get(0));

            return l.get(0);
        }

        System.out.println("Syntax error.");

        return null;
    }

    public List filter(ArrayList topmost, int startSym) {
        List l = new LinkedList();
        for (Iterator i = topmost.iterator(); i.hasNext(); ) {
            Node n = (Node) i.next();
            for (Link link = n.out; link != null; link = link.next) {
                Object sval = link.semAction.run();
                System.out.println("parser result = " + sval);
                l.add(sval);
            }
        }
        return l;
    }

    boolean accepting(Node n, int startSym) {
        return action(actionTable[n.state][parser.eof().symbol()]) == ACCEPT
            && actionData(actionTable[n.state][parser.eof().symbol()]) == startSym;
    }

    // Pre-allocate the return array of reductions.  The parser can,
    // at most, reduce using all rules at once, plus one to indicate the end.
    int[] globalRules;

    // return all reduce actions for state n.state for Terminal t.
    int[] reductions(Node n, Terminal t) {
        int[] rules = globalRules;

        int entry = actionTable[n.state][t.symbol()];
        int action = action(entry);

        if (action == REDUCE) {
            // get the rule to reduce with
            int rule = actionData(entry);
            rules[0] = rule;
            rules[1] = -1;
            return rules;
        }
        if (action == OVERFLOW) {
            // data is an index into the overflowTable
            int start = actionData(entry);

            // get the number of actions in this string of actions.
            int length = overflowTable[start];

            int count = 0;

            for (int i = 0; i < length; i++) {
                entry = overflowTable[start+1+i];
                action = action(entry);
                if (action == REDUCE) {
                    rules[count++] = actionData(entry);
                }
            }

            rules[count] = -1;

            return rules;
        }

        rules[0] = -1;
        return rules;
    }

    // return the shift action for state n.state for Terminal t, if any;
    // -1 otherwise
    int shift(Node n, Terminal t) {
        int entry = actionTable[n.state][t.symbol()];
        int action = action(entry);

        if (action == SHIFT) {
            // data is the destination state
            int dest = actionData(entry);
            return dest;
        }
        if (action == OVERFLOW) {
            // data is an index into the overflowTable
            int start = actionData(entry);

            // get the number of actions in this string of actions.
            int length = overflowTable[start];

            for (int i = 0; i < length; i++) {
                entry = overflowTable[start+1+i];
                if (action(entry) == SHIFT) {
                    return actionData(entry);
                }
            }
        }

        return -1;
    }

    void enqueuePaths(Node root, int depth, int rule,
                      Node useTop, Node useBottom, PathQueue pathqueue) {
        enqueuePathsDFS(root, depth, rule, useTop, useBottom, pathqueue,
                        0, new Path(root, depth));
    }

    void enqueuePathsDFS(Node current, int depth, int rule,
                         Node useTop, Node useBottom,
                         PathQueue pathqueue, int currDepth, Path p)
    {
        if (currDepth == depth) {
            if (useTop == null && useBottom == null) {
                // Add a _copy_ of the path to the queue, since it will
                // be destructively updated for the next path
System.out.println("enqueue path for rule " + rule);
    System.out.print("    " + 0 + ": " + p.top);
for (int i = 0; i < p.path.length; i++) {
    System.out.println(" " + p.path[i].semAction);
    System.out.print(" -> " + (i+1) + ": " + p.path[i].bottom);
}
System.out.println();
                pathqueue.add(new PathQueueEntry(new Path(p), rule));
            }
            return;
        }

        for (Link l = current.out; l != null; l = l.next) {
            // Set useTop, useBottom to null to indicate the link
            // has been seen.
            if (current == useTop && l.bottom == useBottom) {
                useTop = useBottom = null;
            }

            p.path[currDepth] = l;

            enqueuePathsDFS(l.bottom, depth, rule, useTop, useBottom,
                            pathqueue, currDepth+1, p);
        }
    }

    public static class Path {
        Node top;
        Link[] path;
        int span;
        
        Path(Node top, int depth) {
            this.top = top;
            this.path = new Link[depth];
            this.span = 0;
        }

        Path(Path that) {
            this.path = new Link[that.path.length];
            for (int i = 0; i < that.path.length; i++)
                this.path[i] = that.path[i];
            this.top = that.top;

            // Compute the number of tokens spanned by the path.
            span = 0;
            for (int i = 0; i < path.length; i++) {
                span += path[i].span;
            }
        }

        Node bottom() {
            if (path.length == 0) return top;
            return path[path.length-1].bottom;
        }

        int span() {
            return span;
        }

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[" + top);
            for (int i = 0; i < path.length; i++) {
                sb.append("->" + path[i]);
            }
            sb.append(" (spans " + span + ")]");
            return sb.toString();
        }
    }

    public static class PathQueue {
        TreeSet s;

        PathQueue() {
            s = new TreeSet();
        }

        void clear() {
            s.clear();
        }

        boolean isEmpty() {
            return s.isEmpty();
        }

        PathQueueEntry get() {
            PathQueueEntry e = (PathQueueEntry) s.first();
            s.remove(e);
            return e;
        }

        void add(PathQueueEntry e) {
            s.add(e);
        }
    }

    public static class PathQueueEntry implements Comparable {
        Path path;
        int rule;

        PathQueueEntry(Path path, int rule) {
            this.path = path;
            this.rule = rule;
        }

        public int compareTo(Object o) {
            if (o instanceof PathQueueEntry) {
                PathQueueEntry that = (PathQueueEntry) o;

                if (this == that) return 0;

                // To avoid the yield-then-merge problem at least for
                // acyclic grammars, the path queue is sorted according
                // to the following two rules.

                // 1. Reductions that span fewer tokens come first.
                int span1 = this.path.span();
                int span2 = that.path.span();

                if (span1 < span2) return -1;
                if (span1 > span2) return 1;

                // 2. If A -> alpha and B -> beta span the same tokens,
                // then A -> alpha comes first if B ->+ A
                // The ruleTable is constructed so that
                // if B ->+ A, all rules with lhs A come first.
                if (this.rule < that.rule) {
                    // this is possibly derived from that; this is first
                    return -1;
                }
                if (this.rule > that.rule) {
                    // that is possibly derived from this; that is first
                    return 1;
                }

                // As a tie breaker, order the paths lexicographically.
                if (this.path.top.state != that.path.top.state) {
                    return this.path.top.state - that.path.top.state;
                }

                Link[] p1 = this.path.path;
                Link[] p2 = that.path.path;

                if (p1.length != p2.length) {
                    return p1.length - p2.length;
                }

                for (int i = 0; i < p1.length; i++) {
                    if (p1[i].bottom != p2[i].bottom) {
                        return p1[i].bottom.state - p2[i].bottom.state;
                    }
                }

                if (true)
                throw new RuntimeException(this + " and " + that + " are incomparable");

                // Same rule, same number of tokens; this should equal that.
                return 0;
            }

            // We come before any other class.
            return -1;
        }

        public boolean equals(Object o) {
            return this == o;
        }

        public String toString() {
            return "<" + path + "," + rule + ">";
        }
    }

    int[][] decodeActionTable() {
        String[] t = parser.encodedActionTable();
        return (int[][]) new Decoder().decode(t);
    }

    int[] decodeOverflowTable() {
        String[] t = parser.encodedOverflowTable();
        return (int[]) new Decoder().decode(t);
    }

    int[][] decodeGotoTable() {
        String[] t = parser.encodedGotoTable();
        return (int[][]) new Decoder().decode(t);
    }

    int[] decodeRuleTable() {
        String[] t = parser.encodedRuleTable();
        return (int[]) new Decoder().decode(t);
    }

    // The top 3 bits of an action table entry indicate the kind of action.
    // The bottom 29 bits are of an entry e are one of the following:
    // if action(e) == SHIFT, the state to transition to after the shift
    // if action(e) == REDUCE, the rule to reduce
    // if action(e) == OVERFLOW, the entry in the overflowTable
    //                            containing the list of actions
    // 0 otherwise

    // The run-time compiler should inline these methods.
    static int action(int entry) {
        return entry >>> 29;
    }

    static int actionData(int entry) {
        return entry & ~(7 << 29);
    }

    // The top 24 bits of a rule table entry are the lhs nonterminal index.
    // The bottom 8 bits of a rule table entry is the rhs length.
    static int ruleLhsIndex(int entry) {
        return entry >>> 8;
    }

    static int ruleRhsLength(int entry) {
        return entry & 0xff;
    }

    // should be a priority queue
    PathQueue globalPathqueue = new PathQueue();

    void doReductions(Terminal t) {
        // Use a single global pathqueue so we don't have to create an array
        // for each terminal, and so we don't have to keep growing the array.
        PathQueue pathqueue = globalPathqueue;
        pathqueue.clear();

        for (Iterator j = topmost.iterator(); j.hasNext(); ) {
            Node current = (Node) j.next();
System.out.println("in state " + current.state);

            int[] rules = reductions(current, t);

            for (int i = 0; i < rules.length && rules[i] != -1; i++) {
                // reduce N -> alpha
System.out.println("enqueue for reduce with rule " + rules[i]);
                int rhsLength = ruleRhsLength(ruleTable[rules[i]]);
                enqueuePaths(current, rhsLength, rules[i], null, null,
                             pathqueue);
            }
        }

        // don't use an iterator; pathqueue can grow as we iterate
        // through it.
        while (! pathqueue.isEmpty()) {
            PathQueueEntry e = pathqueue.get();
            reduceViaPath(e.path, e.rule, t, pathqueue);
        }
    }

    /** Reset the determinstic depth of all nodes in the GSS reachable
     * from current. */ 
    int resetDeterminsticDepth(Node current) {
        if (current.out == null) {
            current.depth = 1;
            return 1;
        }
        else {
            int m = 0;
            for (Link l = current.out; l != null; l = l.next) {
                int d = resetDeterminsticDepth(l.bottom);
                if (m < d) m = d;
            }
            current.depth = m+1;
            return m+1;
        }
    }

    Link addLink(Node bottom, Node top, Action semAction, int span) {
        Link link = new Link(bottom, top, semAction, span);

        if (top.out == null) {
            top.depth = bottom.depth+1;
        }
        else {
            resetDeterminsticDepth(top);
        }

        // add link to current's list of out links
        link.next = top.out;
        top.out = link;

        // bump the bottom node's refcount
        bottom.refcount++;

        return link;
    }

    /**
     * Return true if accepting, leaving only the accepting states in topmost.
     * Otherwise, do nothing.
     */
    boolean doAccepts(Terminal t, int startSym) {
        List acceptingStates = null;
        for (Iterator i = topmost.iterator(); i.hasNext(); ) {
            Node current = (Node) i.next();
            int e = actionTable[current.state][t.symbol()];

            if (action(e) == ACCEPT && actionData(e) == startSym) {
                System.out.println("accept for " + startSym + " in " + current + " at " + t);
                if (acceptingStates == null) {
                    acceptingStates = new ArrayList();
                }
                acceptingStates.add(current);
            }
        }

        if (acceptingStates != null) {
            topmost.clear();
            topmost.addAll(acceptingStates);
            return true;
        }

        return false;
    }

    void doShifts(Terminal t) {
        ArrayList prevTops = new ArrayList(topmost);

        // "pop" the stack
        topmost.clear();

        // We've done all the reductions for this terminal.
        // The list topmost contains all the states we were in
        // as well as the new states we went to after reducing.
        // Unless there's a parser error at least one of the new
        // states, should shift.
        // We process the old states as well, since there may have been
        // a shift/reduce conflict in one of them and we need to
        // perform the shift there.
        //
        // In the end, any states for which we cannot shift, are syntax
        // errors on the paths being explored and will be removed from
        // topmost.
        //
        // Idea for error recovery: rather than simply removing these
        // states, pop the stack until reaching a state where the shift
        // can be performed, or pop to a error recovery state and don't
        // shift.  By not shifting the input is effectively being
        // discarded for these states until a parser state can be
        // synchronized with the input again.  Of course, if you can't
        // shift an error state, don't pop it; just leave it on the
        // stack.  Error states can be processed as normal, except when
        // reducing from, or through, an error state record the syntax
        // error.
        //
        // ONLY do error recovery, if no states can shift.

        TerminalAction termAction = new TerminalAction(t);

        TOPMOST:
        for (Iterator i = prevTops.iterator(); i.hasNext(); ) {
            Node current = (Node) i.next();
System.out.println("in state " + current.state);

            int dest = shift(current, t);

            if (dest == -1) {
                System.out.println("not shifting " + t);
                // error(current, t);
                continue TOPMOST;
            }

            System.out.println("shifting " + t + " to " + dest);

            // check if there is an existing node topSib in topmost with
            // state dest
            for (Iterator j = topmost.iterator(); j.hasNext(); ) {
                Node topSib = (Node) j.next();

                if (topSib.state == dest) {
                    // add another link, then continue with the next
                    // element of topmost
                    addLink(current, topSib, termAction, 1);
                    continue TOPMOST;
                }
            }

            // no such node; create one and add to topmost
            Node topSib = new Node(dest);
            topmost.add(topSib);

            // push topSib onto stack
            addLink(current, topSib, termAction, 1);
        }

        // Error handling.
        if (topmost.isEmpty()) {
            for (Iterator i = prevTops.iterator(); i.hasNext(); ) {
                Node current = (Node) i.next();
                error(current, t);
            }
        }
    }

    // given a path p in the GSS, which corresponds to an instance of alpha,
    // reduce it to N; t is the current lookahead token; there are three
    // possible outcomes: a new node is created, a new stack link is added
    // between two existing nodes, or a semantic value is merged.
    void reduceViaPath(Path p, int rule, Terminal t, PathQueue pathqueue) {
        // Apply the semantic action for the reduce action.
        // Note that semanticAction expects the semActions array in the
        // same order as the rule rhs; that is,
        // semActions[0] should be the leftmost action.
        // Since, p.path[0].semAction is the semantic action for the rightmost
        // symbol, we need to reverse the list.  symbol on the right-hand-side.
        // Luckily the array was constructed in that order.

        Action[] semActions = new Action[p.path.length];
        for (int i = 0; i < p.path.length; i++) {
            semActions[semActions.length-1-i] = p.path[i].semAction;
        }

        SemanticAction semAction = new SemanticAction(rule, semActions);

        int lhs = ruleLhsIndex(ruleTable[rule]);
        int rhsLength = ruleRhsLength(ruleTable[rule]);

        // let bottomSib be the bottom stack node in p
        Node bottomSib = p.bottom();

        // we want to push the next state, goto(bottomSib,N),
        // onto the stack above bottomSib
        int nextState = gotoTable[bottomSib.state][lhs];

System.out.println("reduce with rule " + rule);
System.out.println("        and goto " + nextState);

        // check if there is already a node topSib with state
        // goto(bottomSib,N) in topmost
        for (Iterator j = topmost.iterator(); j.hasNext(); ) {
            Node topSib = (Node) j.next();

            if (topSib.state == nextState) {
System.out.println("state " + nextState + " already in topmost");

                // check if there is already a link from topSib to bottomSib
                // if there is, merge their semantic values
                for (Link link = topSib.out; link != null; link = link.next) {
                    if (link.bottom == bottomSib) {
                        // there is already a link from topSib to bottomSib
                        // merge the competing interpretations
System.out.println("link from " + topSib + " to " + bottomSib + " already found");
System.out.println("merge(" + lhs + ", " + link.semAction + ", " + semAction + ")");


                        if (link.semAction instanceof TerminalAction) {
                            throw new RuntimeException("Fatal error: cannot merge ambiguous terminals.  How did we get here?");
                        }

                        SemanticAction linkAction = (SemanticAction) link.semAction;

                        // prepend semAction to link's ambiguous list
                        semAction.ambiguous = linkAction.ambiguous;
                        linkAction.ambiguous = semAction;
                        return;
                    }
                }

System.out.println("adding link from " + topSib + " to " + bottomSib);

                // there is no link from topSib to bottomSib
                // we need to create one

                // push topSib onto stack by creating a link
                Link link = addLink(bottomSib, topSib, semAction, p.span());

                // the new link might enable reductions in states whose
                // reductions we've already expanded
                enqueueLimitedReductions(bottomSib, topSib, t, pathqueue);

                return;
            }
        }

        // the next state is not already in topmost; we need to add it.
        System.out.println("pushing " + nextState);

        Node topSib = new Node(nextState);

        // create a link from the new node to the bottom of the stack"
        addLink(bottomSib, topSib, semAction, p.span());

        // insert topSib into topmost;
        topmost.add(topSib);

        // check if the new state can be reduced.  If so, enqueue it
        // the reduction path.
        int[] rules = reductions(topSib, t);

        for (int i = 0; i < rules.length && rules[i] != -1; i++) {
            // reduce N -> alpha
System.out.println("secondary enqueue for reduce with rule " + rules[i]);
            rhsLength = ruleRhsLength(ruleTable[rules[i]]);
            enqueuePaths(topSib, rhsLength, rules[i], null, null, pathqueue);
        }
    }

    void enqueueLimitedReductions(Node bottom, Node top, Terminal t, PathQueue pathqueue) {
        for (Iterator i = topmost.iterator(); i.hasNext(); ) {
            Node n = (Node) i.next();

            int[] rules = reductions(n, t);

            for (int j = 0; j < rules.length && rules[j] != -1; j++) {
                // reduce N -> alpha
System.out.println("secondary (limited) enqueue for reduce using " + rules[j]);
                int rhsLength = ruleRhsLength(ruleTable[rules[j]]);

                // enqueue <p, N -> alpha> for each path p of length |alpha|
                // from n that uses the link top -> bottom
                enqueuePaths(n, rhsLength, rules[j], top, bottom, pathqueue);
            }
        }
    }
}
